%  Copyright (C) 2003-2004 David Roundy
%
%  This program is free software; you can redistribute it and/or modify
%  it under the terms of the GNU General Public License as published by
%  the Free Software Foundation; either version 2, or (at your option)
%  any later version.
%
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License for more details.
%
%  You should have received a copy of the GNU General Public License
%  along with this program; see the file COPYING.  If not, write to
%  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
%  Boston, MA 02110-1301, USA.

\section{Dependencies}
\begin{code}
{-# OPTIONS -fglasgow-exts #-}
#include "gadts.h"
module Darcs.Patch.Depends ( get_common_and_uncommon, get_tags_right,
                 get_common_and_uncommon_or_missing,
                 optimize_patchset, deep_optimize_patchset,
                 slightly_optimize_patchset,
                 get_patches_beyond_tag, get_patches_in_tag,
                 is_tag,
                 patchset_union, patchset_intersection, patchset_complement,
                 commute_to_end,
               ) where
import Data.List ( delete, intersect )
import Control.Monad ( liftM2 )
import Control.Monad.Error (Error(..), MonadError(..))

import Darcs.Patch ( RepoPatch, Named, getdeps, commutex,
                     commuteFL, commuteRL,
                     patch2patchinfo, merge )
import Darcs.Patch.Ordered ( (:\/:)(..), (:<)(..), (:/\:)(..), (:>)(..),
                             unsafeUnRL, unsafeRL,
                             RL(..), FL(..),
                             (+>+), (+<+),
                             reverseFL, mapFL_FL, mapFL,
                             lengthRL, concatRL, reverseRL, mapRL, mapRL_RL )
import Darcs.Patch.Info ( PatchInfo, just_name, human_friendly )
import Darcs.Patch.Set ( PatchSet )
import Darcs.Hopefully ( PatchInfoAnd, piap, info,
                         hopefully, conscientiously, hopefullyM )
import Printer ( errorDoc, renderString, ($$), text )
#include "impossible.h"
\end{code}

\begin{code}
get_common_and_uncommon :: RepoPatch p => (PatchSet p,PatchSet p) ->
                           ([PatchInfo],PatchSet p,PatchSet p)
get_common_and_uncommon_or_missing :: RepoPatch p => (PatchSet p,PatchSet p) ->
                                      Either PatchInfo ([PatchInfo],PatchSet p,PatchSet p)
\end{code}

\begin{code}
get_common_and_uncommon = 
    either missingPatchError id . get_common_and_uncommon_err

get_common_and_uncommon_or_missing = 
    either (\(MissingPatch x _) -> Left x) Right . get_common_and_uncommon_err

get_common_and_uncommon_err :: RepoPatch p => (PatchSet p,PatchSet p) ->
                               Either MissingPatch ([PatchInfo],PatchSet p,PatchSet p)
get_common_and_uncommon_err (ps1,ps2) = gcau (optimize_patchset ps1) ps2

{-
with_partial_intersection takes two PatchSets and splits them into a common
intersection portion and two sets of patches.  The intersection, however,
is only lazily determined, so there is no guarantee that all intersecting
patches will be included in the intersection PatchSet.  This is a pretty
efficient function, because it makes use of the already-broken-up nature of
PatchSets.

PatchSets have the property that if

(info $ last $ head a) == (info $ last $ head b)

then (tail a) and (tail b) are identical repositories, and we want to take
advantage of this if possible, to avoid reading too many inventories.  In
the case of --partial repositories or patch bundles, it is crucial that we
don't need to read the whole history, since it isn't available.

TODO:

The length equalising isn't necessarily right. We probably also be
thinking about not going past the end of a partial repository, or favour
local repository stuff over remote repository stuff.

Also, when comparing l1 to l2, we should really be comparing the
newly discovered one to /all/ the lasts in the other patch set
that we've got so far.
-}

with_partial_intersection :: RepoPatch p => PatchSet p C(x) -> PatchSet p C(y)
                          -> (FORALL(z) PatchSet p C(z) -> RL (PatchInfoAnd p) C(z x)
                                                      -> RL (PatchInfoAnd p) C(z y) -> a)
                          -> a
with_partial_intersection NilRL ps2 j = j (NilRL:<:NilRL) NilRL (concatRL ps2)
with_partial_intersection ps1 NilRL j = j (NilRL:<:NilRL) (concatRL ps1) NilRL
with_partial_intersection (NilRL:<:ps1) ps2 j =
    with_partial_intersection ps1 ps2 j
with_partial_intersection ps1 (NilRL:<:ps2) j =
    with_partial_intersection ps1 ps2 j
-- NOTE: symmetry is broken here, so we want the PatchSet with more history
-- first!
with_partial_intersection ((pi1:<:NilRL):<:common) ((pi2:<:NilRL):<:_) j
    | info pi1 == info pi2 = j common NilRL NilRL
with_partial_intersection (orig_ps1:<:orig_ps1s) (orig_ps2:<:orig_ps2s) j
 = f (lengthRL orig_ps1) (last $ mapRL info orig_ps1) (orig_ps1:>:NilFL) orig_ps1s
     (lengthRL orig_ps2) (last $ mapRL info orig_ps2) (orig_ps2:>:NilFL) orig_ps2s
    where {- Invariants: nx = length $ cr psx
                         lx = last $ cr psx   -}
          f _n1 l1 ps1 ps1s _n2 l2 ps2 _ps2s
           | l1 == l2 = j ps1s (cr ps1) (cr ps2)
          f n1 l1 ps1 ps1s n2 l2 ps2 ps2s
           = case compare n1 n2 of
             GT -> case dropWhileNilRL ps2s of
                   ps2':<:ps2s' ->
                       f n1 l1 ps1 ps1s
                         (n2 + lengthRL ps2') (last $ mapRL info ps2') (ps2':>:ps2) ps2s'
                   NilRL -> -- We keep going round f so the l1 == l2 case
                            -- has a chance to kick in
                         case dropWhileNilRL ps1s of
                         ps1':<:ps1s' ->
                             f (n1 + lengthRL ps1') (last $ mapRL info ps1')
                               (ps1':>:ps1) ps1s'
                               n2 l2 ps2 ps2s
                         NilRL -> j (NilRL:<:NilRL) (cr ps1) (cr ps2)
             _  -> case dropWhileNilRL ps1s of
                   ps1':<:ps1s' ->
                       f (n1 + lengthRL ps1') (last $ mapRL info ps1') (ps1':>:ps1) ps1s'
                         n2 l2 ps2 ps2s
                   NilRL -> -- We keep going round f so the l1 == l2 case
                            -- has a chance to kick in
                         case dropWhileNilRL ps2s of
                         ps2':<:ps2s' ->
                             f n1 l1 ps1 NilRL
                               (n2 + lengthRL ps2') (last $ mapRL info ps2')
                               (ps2':>:ps2) ps2s'
                         NilRL -> j (NilRL:<:NilRL) (cr ps1) (cr ps2)
          cr :: FL (RL a) -> RL a
          cr = concatRL . reverseFL

{-
gcau determines a list of "common" patches and patches unique to each of
the two PatchSets.  The list of "common" patches only needs to include all
patches that are not interspersed with the "unique" patches, but including
more patches in the list of "common" patches doesn't really hurt, except
for efficiency considerations.  Mostly, we want to access as few elements
as possible of the PatchSet list, since those can be expensive (or
unavailable).

PatchSets have the property that if

(info $ last $ head a) == (info $ last $ head b)

then (tail a) and (tail b) are identical repositories, and we want to take
advantage of this if possible, to avoid reading too many inventories.  In
the case of --partial repositories or patch bundles, it is crucial that we
don't need to read the whole history, since it isn't available.

TODO:

The length equalising isn't necessarily right. We probably also be
thinking about not going past the end of a partial repository, or favour
local repository stuff over remote repo stuff.

Also, when comparing l1 to l2, we should really be comparing the
newly discovered one to /all/ the lasts in the other patch set
that we've got so far.
-}


gcau :: RepoPatch p => PatchSet p -> PatchSet p
     -> Either MissingPatch ([PatchInfo],PatchSet p,PatchSet p)
gcau NilRL ps2 = return ([], NilRL:<:NilRL, concatRL ps2 :<: NilRL)
gcau ps1 NilRL = return ([], concatRL ps1 :<: NilRL, NilRL:<:NilRL)
gcau (NilRL:<:ps1) ps2 = gcau ps1 ps2
gcau ps1 (NilRL:<:ps2) = gcau ps1 ps2
gcau ((pi1:<:NilRL):<:_) ((pi2:<:NilRL):<:_)
 | info pi1 == info pi2 = return ([info pi1], NilRL:<:NilRL, NilRL:<:NilRL)
gcau (orig_ps1:<:orig_ps1s) (orig_ps2:<:orig_ps2s)
 = f (lengthRL orig_ps1) (last $ mapRL info orig_ps1) (orig_ps1:>:NilFL) orig_ps1s
     (lengthRL orig_ps2) (last $ mapRL info orig_ps2) (orig_ps2:>:NilFL) orig_ps2s
    where {- Invariants: nx = lengthRL $ cr psx
                         lx = last $ cr psx   -}
          f :: RepoPatch p => Int -> PatchInfo -> FL (RL (PatchInfoAnd p)) -> PatchSet p
            -> Int -> PatchInfo -> FL (RL (PatchInfoAnd p)) -> PatchSet p
            -> Either MissingPatch ([PatchInfo],PatchSet p,PatchSet p)
          f _n1 l1 ps1 _ps1s _n2 l2 ps2 _ps2s
           | l1 == l2 = gcau_simple (cr ps1) (cr ps2)
          f n1 l1 ps1 ps1s n2 l2 ps2 ps2s
           = case compare n1 n2 of
             GT -> case dropWhileNilRL ps2s of
                   ps2':<:ps2s' ->
                       f n1 l1 ps1 ps1s
                         (n2 + lengthRL ps2') (last $ mapRL info ps2') (ps2':>:ps2) ps2s'
                   NilRL -> -- We keep going round f so the l1 == l2 case
                            -- has a chance to kick in
                         case dropWhileNilRL ps1s of
                         ps1':<:ps1s' ->
                             f (n1 + lengthRL ps1') (last $ mapRL info ps1')
                               (ps1':>:ps1) ps1s'
                               n2 l2 ps2 ps2s
                         NilRL -> gcau_simple (cr ps1) (cr ps2)
             _  -> case dropWhileNilRL ps1s of
                   ps1':<:ps1s' ->
                       f (n1 + lengthRL ps1') (last $ mapRL info ps1') (ps1':>:ps1) ps1s'
                         n2 l2 ps2 ps2s
                   NilRL -> -- We keep going round f so the l1 == l2 case
                            -- has a chance to kick in
                         case dropWhileNilRL ps2s of
                         ps2':<:ps2s' ->
                             f n1 l1 ps1 NilRL
                               (n2 + lengthRL ps2') (last $ mapRL info ps2')
                               (ps2':>:ps2) ps2s'
                         NilRL -> gcau_simple (cr ps1) (cr ps2)
          cr :: FL (RL a) -> RL a
          cr = concatRL . reverseFL

dropWhileNilRL :: PatchSet p C(x) -> PatchSet p C(x)
dropWhileNilRL (NilRL:<:xs) = dropWhileNilRL xs
dropWhileNilRL xs = xs

gcau_simple :: RepoPatch p => RL (PatchInfoAnd p) -> RL (PatchInfoAnd p)
            -> Either MissingPatch ([PatchInfo],PatchSet p,PatchSet p)
gcau_simple ps1 ps2 = do
 ex1 <- get_extra (return NilFL) common ps1
 ex2 <- get_extra (return NilFL) common ps2
 -- Note: Hairy stuff going on here, which is only okay because we ignore the
 -- actual values of these patches!
 let ps1' = filter ((`elem` common) . info) $ unsafeUnRL ps1
 return (mapRL info $ headRL $
         ((optimize_patchset $ unsafeRL ps1' :<: NilRL) +<+ (NilRL:<:NilRL)) , ex1 :<: NilRL , ex2 :<: NilRL)
  where common = mapRL info ps1 `intersect` mapRL info ps2
        headRL (x:<:_) = x
        headRL _ = impossible

data MissingPatch = MissingPatch !PatchInfo !String

instance Error MissingPatch where
    -- we don't really need those methods
    noMsg = bug "MissingPatch doesn't define noMsg."

get_extra :: RepoPatch p => Either MissingPatch (FL (Named p))
          -> [PatchInfo]
          -> RL (PatchInfoAnd p)
          -> Either MissingPatch (RL (PatchInfoAnd p))
get_extra _ _ NilRL = return NilRL
get_extra skipped common (hp:<:pps) =
    if info hp `elem` common && is_tag (info hp)
    then case getdeps `fmap` hopefullyM hp of
         Just ds -> get_extra (liftM2 (:>:) ep skipped) (ds++delete (info hp) common) pps
         Nothing -> get_extra (liftM2 (:>:) ep skipped) (delete (info hp) common) pps
    else if info hp `elem` common
         then get_extra (liftM2 (:>:) ep skipped) (delete (info hp) common) pps
         else do
            p <- ep
            skpd <- skipped
            case commuteFL (p :> skpd) of
              Just (skipped_patch' :> p') -> do
                  x <- get_extra (return skipped_patch') common pps
                  return (info hp `piap` p' :<: x)
              Nothing -> errorDoc $ text "bug in get_extra commuting patch:"
                         $$ human_friendly (info hp)
    where ep = case hopefullyM hp of
               Right p' -> return p'
               Left e -> throwError (MissingPatch (info hp) e)

missingPatchError :: MissingPatch -> a
missingPatchError (MissingPatch pinfo e) =
    errorDoc
        ( text "failed to read patch in get_extra:"
          $$ human_friendly pinfo $$ text e
          $$ text "Perhaps this is a 'partial' repository?" )

get_extra_old :: RepoPatch p => FL (Named p)
              -> [PatchInfo]
              -> RL (PatchInfoAnd p)
              -> RL (PatchInfoAnd p)
get_extra_old skipped common pps =
    either missingPatchError id (get_extra (return skipped) common pps)
              
\end{code}

\begin{code}
get_patches_beyond_tag :: RepoPatch p => PatchInfo -> PatchSet p -> PatchSet p
get_patches_beyond_tag t ((hp:<:NilRL):<:_) | info hp == t = NilRL :<: NilRL
get_patches_beyond_tag t patchset@((hp:<:ps):<:pps) =
    if info hp == t
    then if get_tags_right patchset == [info hp]
         then NilRL :<: NilRL -- special case to avoid looking at redundant patches
         else get_extra_old NilFL [t] (concatRL patchset) :<: NilRL
    else hp -:- get_patches_beyond_tag t (ps:<:pps)
get_patches_beyond_tag t (NilRL:<:pps) = get_patches_beyond_tag t pps
get_patches_beyond_tag t NilRL = bug $ "tag\n" ++
                                 renderString (human_friendly t) ++
                                 "\nis not in the patchset in get_patches_beyond_tag."

get_patches_in_tag :: RepoPatch p => PatchInfo -> PatchSet p -> PatchSet p
get_patches_in_tag t pps@((hp:<:NilRL):<:xs)
    | info hp == t = pps
    | otherwise = get_patches_in_tag t xs

get_patches_in_tag t ((hp:<:ps):<:xs)
    | info hp /= t = get_patches_in_tag t (ps:<:xs)

get_patches_in_tag t ((pa:<:ps):<:xs) = gpit thepis (pa:>:NilFL) (ps:<:xs)
    where thepis = getdeps $ conscientiously
                   (\e -> text "Couldn't read tag"
                          $$ human_friendly t
                          $$ text ""
                          $$ e) pa
          gpit _ sofar NilRL = reverseFL sofar :<: NilRL
          gpit deps sofar ((hp:<:NilRL):<:xs')
              | info hp `elem` deps = (reverseFL $ hp :>: sofar) :<: xs'
              | otherwise = gpit deps sofar xs'
          gpit deps sofar (NilRL:<:xs') = gpit deps sofar xs'
          gpit deps sofar ((hp:<:ps'):<:xs')
              | info hp `elem` deps
                  = let odeps = filter (/=info hp) deps
                        alldeps = if is_tag $ info hp
                                  then odeps ++ getdeps (hopefully hp)
                                  else odeps
                    in gpit alldeps (hp:>:sofar) (ps':<:xs')
              | otherwise
                  = gpit deps (commute_by sofar $ hopefully hp) (ps':<:xs')

get_patches_in_tag t _ = errorDoc $ text "Couldn't read tag"
                                 $$ human_friendly t

commute_by :: RepoPatch p => FL (PatchInfoAnd p) C(x y) -> (Named p) C(w x)
           -> FL (PatchInfoAnd p) C(w z)
commute_by NilFL _ = NilFL
commute_by (hpa:>:xs) p =
    case commutex (hopefully hpa :< p) of
    Nothing -> bug "Failure commuting patches in commute_by called by gpit!"
    Just (p':<a') -> (info hpa `piap` a') :>: commute_by xs p'
\end{code}

\begin{code}
is_tag :: PatchInfo -> Bool
is_tag pinfo = take 4 (just_name pinfo) == "TAG "

get_tags_right :: PatchSet p -> [PatchInfo]
get_tags_right NilRL = []
get_tags_right (ps:<:_) = get_tags_r ps
    where
    get_tags_r NilRL = []
    get_tags_r (hp:<:pps)
        | is_tag (info hp) = case getdeps `fmap` hopefullyM hp of
                             Just ds -> info hp : get_tags_r (drop_tags_r ds pps)
                             Nothing -> info hp : mapRL info pps
        | otherwise = info hp : get_tags_r pps
    drop_tags_r :: [PatchInfo]
                -> RL (PatchInfoAnd p) -> RL (PatchInfoAnd p)
    drop_tags_r [] pps = pps
    drop_tags_r _ NilRL = NilRL
    drop_tags_r ds (hp:<:pps)
        | info hp `elem` ds && is_tag (info hp) =
            case getdeps `fmap` hopefullyM hp of
            Just ds' -> drop_tags_r (ds'++delete (info hp) ds) pps
            Nothing -> drop_tags_r (delete (info hp) ds) pps
        | info hp `elem` ds = drop_tags_r (delete (info hp) ds) pps
        | otherwise = hp :<: drop_tags_r ds pps
\end{code}

\begin{code}
deep_optimize_patchset :: PatchSet p -> PatchSet p
deep_optimize_patchset pss = optimize_patchset (concatRL pss :<: NilRL)

optimize_patchset :: PatchSet p -> PatchSet p
optimize_patchset NilRL = NilRL
optimize_patchset (ps:<:pss) = opsp ps +<+ pss
opsp :: RL (PatchInfoAnd p) -> PatchSet p
opsp NilRL = NilRL
opsp (hp:<:pps)
     | is_tag (info hp) && get_tags_right ((hp:<:pps):<:NilRL) == [info hp]
         = (hp:<:NilRL) :<: opsp pps
     | otherwise = hp -:- opsp pps

(-:-) :: (PatchInfoAnd p) -> PatchSet p -> PatchSet p
pp -:- NilRL = (pp:<:NilRL) :<: NilRL
pp -:- (p:<:ps) = ((pp:<:p) :<: ps)

slightly_optimize_patchset :: PatchSet p -> PatchSet p
slightly_optimize_patchset NilRL = NilRL
slightly_optimize_patchset (ps:<:pss) = sops ps +<+ pss
    where sops NilRL = NilRL
          sops (pinfomp :<: NilRL) = (pinfomp :<: NilRL) :<: NilRL
          sops (hp:<:pps) | is_tag (info hp) &&
                            get_tags_right ((hp:<:pps):<:NilRL) == [info hp]
                                = (hp:<:NilRL) :<: (pps:<: NilRL)
                          | otherwise = hp -:- sops pps
\end{code}

\begin{code}
commute_to_end :: RepoPatch p => FL (Named p) -> PatchSet p
               -> (FL (Named p) :< FL (Named p))
commute_to_end select from =
   ctt NilFL (mapFL patch2patchinfo select) (concatRL from)
   where
      ctt :: RepoPatch p => FL (Named p) -> [PatchInfo]
          -> RL (PatchInfoAnd p) -> (FL (Named p) :< FL (Named p))
      ctt skp [] _ = (NilFL :< skp)
      ctt skp sel (hp:<:ps)
         | info hp `elem` sel
            = case commuteFL (hopefully hp :> skp) of
              Nothing -> bug "patches to commute_to_end does not commutex (1)"
              Just (skp' :> p') ->
                 let (ps' :< skp'') = ctt skp' (delete (info hp) sel) ps
                 in ((p':>:ps') :< skp'')
         | otherwise
            = ctt (hopefully hp:>:skp) sel ps
      ctt _ _ _ = bug "patches to commute_to_end does not commutex (2)"

segregate_patches :: RepoPatch p => (PatchInfo -> Bool) -> PatchSet p C(z)
                  -> (RL (Named p) :< RL (Named p)) C(() z)
segregate_patches select_first from =
   ctt NilFL NilFL (concatRL from)
   where
      ctt :: RepoPatch p => FL (Named p)  C(y z) -> FL (Named p) C(x y) -> RL (PatchInfoAnd p) C(o x)
          -> (RL (Named p) :< RL (Named p)) C(o z)
      ctt las fir NilRL = reverseFL las :< reverseFL fir
      ctt las fir (hp:<:ps)
         | select_first (info hp) = ctt las (hopefully hp:>:fir) ps
         | otherwise =
             case commuteFL (hopefully hp :> fir) of
             Nothing -> bug "patches to segregate_patches does not commutex (1)"
             Just (fir' :> p') -> ctt (p':>:las) fir' ps
\end{code}

\begin{code}
patchset_intersection :: RepoPatch p => [PatchSet p] -> PatchSet p
patchset_intersection [] = NilRL :<: NilRL
patchset_intersection [x] = x
patchset_intersection (y:ys) =
    with_partial_intersection y (patchset_intersection ys) $
    \common a b ->        
        let morecommon = mapRL info a `intersect` mapRL info b
            _ :< commonps = segregate_patches (`elem` morecommon) (a :<: NilRL)
        in
        (mapRL_RL (\p -> (patch2patchinfo p `piap` p)) commonps) :<: common

patchset_union :: RepoPatch p => [PatchSet p] -> PatchSet p
patchset_union [] = NilRL :<: NilRL
patchset_union [x] = x
patchset_union (y:ys) =
    with_partial_intersection y (patchset_union ys) $
    \common a b ->
        case gcau_simple a b of
        Left e -> missingPatchError e
        Right (_, a' :<: NilRL, b' :<: NilRL) -> (merge_sets a' b' +<+ b) :<: common
        _ -> impossible

merge_sets :: RepoPatch p => RL (PatchInfoAnd p) -> RL (PatchInfoAnd p) -> RL (PatchInfoAnd p)
merge_sets l r =
    let pl = mapFL_FL hopefully $ reverseRL l
        pr = mapFL_FL hopefully $ reverseRL r
        p2pimp p = patch2patchinfo p `piap` p
    in case merge (pl:\/: pr) of
       (_:/\:pl') -> reverseFL $ mapFL_FL p2pimp pl'
\end{code}

\begin{code}
patchset_complement :: RepoPatch p => PatchSet p C(x) -> PatchSet p C(y) -> PatchSet p
patchset_complement (NilRL :<: NilRL) _ = NilRL :<: NilRL
patchset_complement x (NilRL :<: NilRL) = x
patchset_complement x1 xs =
    with_partial_intersection x1 (patchset_union [xs]) $
        \_common a b -> let common' :: [PatchInfo]
                            common' = (mapRL info a) `intersect` (mapRL info b)
                            {- need to commutex common' to the end of candidates, since common' will not be
                               pulled and other candidates therefore cannot depend on common'...  anything
                               which depends on common' will be pushed to the end along with common'. The
                               push_to_end function is similar to get_extra except that it commutes in
                               the opposite direction: pushing common to be the last patches, moving "extras"
                               to be applied before common (and common's deps).  -}
                            targets = push_to_end (Right NilFL) (Right NilRL) common' $ reverseRL a
                            push_to_end :: RepoPatch p => Either MissingPatch (FL (PatchInfoAnd p))
                                           -> Either MissingPatch (RL (Named p))
                                           -> [PatchInfo]
                                           -> FL (PatchInfoAnd p)
                                           -> Either MissingPatch (FL (PatchInfoAnd p))
                            push_to_end before _ _ NilFL = before
                            push_to_end before pushing pushcore (hp:>:patchset) = 
                                let ep = either (throwError . MissingPatch (info hp)) Right
                                         $ hopefullyM hp
                                    pushing_add = do ps <- pushing
                                                     thep <- ep
                                                     return (thep :<: ps)
                                    before_add i = (+>+ ((info hp `piap` i) :>: NilFL)) `fmap` before
                                    push_patch = push_to_end before pushing_add (delete (info hp) pushcore) patchset
                                    indep_patch (indep:>push') = push_to_end (before_add indep) (return push') pushcore patchset
                                in if info hp `elem` pushcore
                                   then push_patch
                                   else do p <- ep
                                           pushing >>= maybe push_patch indep_patch . commuteRL . (:> p)
                            ind_ps :: FL a -> RL (RL  a)
                            ind_ps  = listof . reverseFL
                            listof :: a -> RL a
                            listof = (:<: NilRL)
                        in  either missingPatchError ind_ps targets
\end{code}
