%  Copyright (C) 2002-2004 David Roundy
%
%  This program is free software; you can redistribute it and/or modify
%  it under the terms of the GNU General Public License as published by
%  the Free Software Foundation; either version 2, or (at your option)
%  any later version.
%
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License for more details.
%
%  You should have received a copy of the GNU General Public License
%  along with this program; see the file COPYING.  If not, write to
%  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
%  Boston, MA 02110-1301, USA.


\begin{code}
{-# OPTIONS -fno-warn-orphans #-}
module Darcs.Patch.Viewing ( xml_summary, summarize )
             where

import Prelude hiding ( pi )
import Control.Monad ( liftM )
import Data.List ( sort )

import Darcs.SlurpDirectory ( Slurpy, get_slurp, get_filecontents )
import FastPackedString ( nullPS )
import FileName ( FileName, fp2fn, fn2fp )
import Printer ( Doc, empty, vcat,
                 text, blueText,
                 minus, plus, ($$), (<+>), (<>),
                 prefix, renderString,
                 userchunkPS,
               )
import Darcs.Patch.Core ( Patch(..), Named(..),
                          patchcontents )
import Darcs.Patch.Prim ( Prim(..), is_hunk, fn2d, showPrim,
                          DirPatchType(..), FilePatchType(..) )
import Darcs.Patch.Effect ( Effect, effect )
import Darcs.Patch.Patchy ( Patchy, Apply, ShowPatch(..), identity )
import Darcs.Patch.Show ( showPatch_, showNamedPrefix )
import Darcs.Patch.Info ( showPatchInfo, human_friendly )
import Darcs.Patch.Apply ( apply_to_slurpy )
#include "impossible.h"
#include "gadts.h"
import Darcs.Patch.Ordered ( RL(..), FL(..),
                             mapFL, mapFL_FL, reverseRL )
\end{code}


\begin{code}
instance ShowPatch Prim where
    showPatch = showPrim
    showContextPatch s p@(FP _ (Hunk _ _ _)) = showContextHunk s (PP p)
    showContextPatch s (Split ps) =
        blueText "(" $$ showContextSeries s (mapFL_FL PP ps)
                     <> blueText ")"
    showContextPatch _ p = showPatch p
    summary = gen_summary False . (:>:NilFL)
    thing _ = "change"

summarize :: Effect e => e C(x y) -> Doc
summarize = gen_summary False . effect

instance ShowPatch Patch where
    showPatch = showPatch_
    showContextPatch s (PP x) | is_hunk x = showContextHunk s (PP x)
    showContextPatch _ (ComP NilFL) = blueText "{" $$ blueText "}"
    showContextPatch s (ComP ps) = blueText "{" $$ showContextSeries s ps
                                   $$ blueText "}"
    showContextPatch _ p = showPatch p
    summary = summarize
    thing _ = "change"
\end{code}

\begin{code}
showContextSeries :: Slurpy -> FL Patch C(x y) -> Doc
showContextSeries slur patches = scs slur identity patches
    where scs :: Slurpy -> Patch C(w x) -> FL Patch C(x y) -> Doc
          scs s pold (PP p:>:p2:>:ps)
              | is_hunk p = coolContextHunk s pold (PP p) p2
                         $$ scs (fromJust $ apply_to_slurpy p s)
                                (PP p) (p2:>:ps)
          scs s pold (PP p:>:NilFL)
              | is_hunk p = coolContextHunk s pold (PP p) identity
          scs s _ (p:>:ps) = showContextPatch s p
                        $$ scs (fromJust $ apply_to_slurpy p s) p ps
          scs _ _ NilFL = empty
\end{code}

\begin{code}
showContextHunk :: Slurpy -> Patch C(x y) -> Doc
showContextHunk s p = coolContextHunk s identity p identity

coolContextHunk :: Slurpy -> Patch C(a b) -> Patch C(b c)
                -> Patch C(c d) -> Doc
coolContextHunk s prev p@(PP (FP f (Hunk l o n))) next =
    case (fst . get_filecontents) `liftM` get_slurp f s of
    Nothing -> showPatch p -- This is a weird error...
    Just ls ->
        let numpre = case prev of
                     (PP (FP f' (Hunk lprev _ nprev)))
                         | f' == f &&
                           l - (lprev + length nprev + 3) < 3 &&
                           lprev < l ->
                             max 0 $ l - (lprev + length nprev + 3)
                     _ -> if l >= 4 then 3 else l - 1
            pre = take numpre $ drop (l - numpre - 1) ls
            numpost = case next of
                      (PP (FP f' (Hunk lnext _ _)))
                          | f' == f && lnext < l+length n+4 &&
                            lnext > l ->
                              lnext - (l+length n)
                      _ -> 3
            cleanedls = case reverse ls of
                        (x:xs) | nullPS x -> reverse xs
                        _ -> ls
            post = take numpost $ drop (max 0 $ l+length o-1) cleanedls
            in blueText "hunk" <+> fn2d f <+> text (show l)
            $$ prefix " " (vcat $ map userchunkPS pre)
            $$ prefix "-" (vcat $ map userchunkPS o)
            $$ prefix "+" (vcat $ map userchunkPS n)
            $$ prefix " " (vcat $ map userchunkPS post)
coolContextHunk _ _ _ _ = impossible
\end{code}

\begin{code}
xml_summary :: (Effect p, Patchy p) => Named p C(x y) -> Doc
xml_summary p = text "<summary>"
             $$ gen_summary True (effect $ patchcontents p)
             $$ text "</summary>"

-- Yuck duplicated code below...
escapeXML :: String -> Doc
escapeXML = text . strReplace '\'' "&apos;" . strReplace '"' "&quot;" .
  strReplace '>' "&gt;" . strReplace '<' "&lt;" . strReplace '&' "&amp;"

strReplace :: Char -> String -> String -> String
strReplace _ _ [] = []
strReplace x y (z:zs)
  | x == z    = y ++ (strReplace x y zs)
  | otherwise = z : (strReplace x y zs)
-- end yuck duplicated code.

gen_summary :: Bool -> FL Prim C(x y) -> Doc
gen_summary use_xml p
    = vcat themoves
   $$ vcat themods
    where themods = map summ $ combine $ sort $ concat $ mapFL s p
          s :: Prim C(x y) -> [(FileName, Int, Int, Int, Bool, Bool)]
          s (FP f (Hunk _ o n)) = [(f, length o, length n, 0, False, False)]
          s (FP f (Binary _ _)) = [(f, 0, 0, 0, False, False)]
          s (FP f AddFile) = [(f, -1, 0, 0, False, False)]
          s (FP f RmFile) = [(f, 0, -1, 0, False, False)]
          s (FP f (TokReplace _ _ _)) = [(f, 0, 0, 1, False, False)]
          s (DP d AddDir) = [(d, -1, 0, 0, True, False)]
          s (DP d RmDir) = [(d, 0, -1, 0, True, False)]
          s (Split xs) = concat $ mapFL s xs
          s (Move _ _) = [(fp2fn "", 0, 0, 0, False, False)]
          s (ChangePref _ _ _) = [(fp2fn "", 0, 0, 0, False, False)]
          s Identity = [(fp2fn "", 0, 0, 0, False, False)]
          (-1) .+ _ = -1
          _ .+ (-1) = -1
          a .+ b = a + b
          combine ((f,a,b,r,isd,c):(f',a',b',r',_,c'):ss)
              -- Don't combine AddFile and RmFile: (maybe an old revision of) darcs
              -- allows a single patch to add and remove the same file, see issue 185
              | f == f' && (a /= -1 || b' /= -1) && (a' /= -1 || b /= -1) = combine ((f,a.+a',b.+b',r+r',isd,c||c'):ss)
          combine ((f,a,b,r,isd,c):ss) = (f,a,b,r,isd,c) : combine ss
          combine [] = []

          summ (f,_,-1,_,False,False)
              = if use_xml then text "<remove_file>"
                             $$ escapeXML (drop_dotslash $ fn2fp f)
                             $$ text "</remove_file>"
                           else text "R" <+> text (fn2fp f)
          summ (f,_,-1,_,False,True)
              = if use_xml then text "<remove_file conflict='true'>"
                             $$ escapeXML (drop_dotslash $ fn2fp f)
                             $$ text "</remove_file>"
                           else text "R!" <+> text (fn2fp f)
          summ (f,-1,_,_,False,False)
              = if use_xml then text "<add_file>"
                             $$ escapeXML (drop_dotslash $ fn2fp f)
                             $$ text "</add_file>"
                           else text "A" <+> text (fn2fp f)
          summ (f,-1,_,_,False,True)
              = if use_xml then text "<add_file conflict='true'>"
                             $$ escapeXML (drop_dotslash $ fn2fp f)
                             $$ text "</add_file>"
                           else text "A!" <+> text (fn2fp f)
          summ (f,0,0,0,False,False) | f == fp2fn "" = empty
          summ (f,0,0,0,False,True) | f == fp2fn ""
              = if use_xml then empty -- don't know what to do here...
                           else text "!" <+> text (fn2fp f)
          summ (f,a,b,r,False,False)
              = if use_xml then text "<modify_file>"
                             $$ escapeXML (drop_dotslash $ fn2fp f)
                                    <> xrm a <> xad b <> xrp r
                             $$ text "</modify_file>"
                           else text "M" <+> text (fn2fp f)
                                         <+> rm a <+> ad b <+> rp r
          summ (f,a,b,r,False,True)
              = if use_xml then text "<modify_file conflict='true'>"
                             $$ escapeXML (drop_dotslash $ fn2fp f)
                                    <> xrm a <> xad b <> xrp r
                             $$ text "</modify_file>"
                           else text "M!" <+> text (fn2fp f)
                                    <+> rm a <+> ad b <+> rp r
          summ (f,_,-1,_,True,False)
              = if use_xml then text "<remove_directory>"
                             $$ escapeXML (drop_dotslash $ fn2fp f)
                             $$ text "</remove_directory>"
                           else text "R" <+> text (fn2fp f) <> text "/"
          summ (f,_,-1,_,True,True)
              = if use_xml then text "<remove_directory conflict='true'>"
                             $$ escapeXML (drop_dotslash $ fn2fp f)
                             $$ text "</remove_directory>"
                           else text "R!" <+> text (fn2fp f) <> text "/"
          summ (f,-1,_,_,True,False)
              = if use_xml then text "<add_directory>"
                             $$ escapeXML (drop_dotslash $ fn2fp f)
                             $$ text "</add_directory>"
                           else text "A" <+> text (fn2fp f) <> text "/"
          summ (f,-1,_,_,True,True)
              = if use_xml then text "<add_directory conflict='true'>"
                             $$ escapeXML (drop_dotslash $ fn2fp f)
                             $$ text "</add_directory>"
                           else text "A!" <+> text (fn2fp f) <> text "/"
          summ _ = empty
          ad 0 = empty
          ad a = plus <> text (show a)
          xad 0 = empty
          xad a = text "<added_lines num='" <> text (show a) <> text "'/>"
          rm 0 = empty
          rm a = minus <> text (show a)
          xrm 0 = empty
          xrm a = text "<removed_lines num='" <> text (show a) <> text "'/>"
          rp 0 = empty
          rp a = text "r" <> text (show a)
          xrp 0 = empty
          xrp a = text "<replaced_tokens num='" <> text (show a) <> text "'/>"
          drop_dotslash ('.':'/':str) = drop_dotslash str
          drop_dotslash str = str
          themoves :: [Doc]
          themoves = mapFL showmoves p
          showmoves :: Prim C(x y) -> Doc
          showmoves (Move a b)
              = if use_xml
                then text "<move from=\""
                  <> escapeXML (drop_dotslash $ fn2fp a) <> text "\" to=\""
                  <> escapeXML (drop_dotslash $ fn2fp b) <> text"\"/>"
                else text " "    <> text (fn2fp a)
                  <> text " -> " <> text (fn2fp b)
          showmoves _ = empty
\end{code}

\begin{code}
instance (Effect p, ShowPatch p) => ShowPatch (Named p) where
    showPatch (NamedP n [] p) = showPatchInfo n <> showPatch p
    showPatch (NamedP n d p) = showNamedPrefix n d <+> showPatch p
    showContextPatch s (NamedP n [] p) = showPatchInfo n <> showContextPatch s p
    showContextPatch s (NamedP n d p) = showNamedPrefix n d <+> showContextPatch s p
    description (NamedP n _ _) = human_friendly n
    summary p = description p $$ text "" $$
                prefix "    " (summarize p)
    showNicely p@(NamedP _ _ pt) = description p $$
                                   prefix "    " (showNicely pt)

instance (Effect p, ShowPatch p) => Show (Named p C(x y)) where
    show = renderString . showPatch

instance (Apply p, ShowPatch p) => ShowPatch (FL p) where
    showPatch xs = vcat (mapFL showPatch xs)
    showContextPatch _ NilFL = empty
    showContextPatch s (p:>:ps) = showContextPatch s p $$
                                  showContextPatch (fromJust $ apply_to_slurpy p s) ps
    description = vcat . mapFL description
    summary = vcat . mapFL summary
    thing x = thing (helperx x) ++ "s"
        where helperx :: FL a C(x y) -> a C(x y)
              helperx _ = undefined
    things = thing

instance (Apply p, ShowPatch p) => ShowPatch (RL p) where
    showPatch = showPatch . reverseRL
    showContextPatch s = showContextPatch s . reverseRL
    description = description . reverseRL
    summary = summary . reverseRL
    thing = thing . reverseRL
    things = things . reverseRL

instance Patchy p => Patchy (FL p)
instance Patchy p => Patchy (RL p)

\end{code}
