%  Copyright (C) 2002-2004 David Roundy
%  Copyright (C) 2005 Juliusz Chroboczek
%
%  This program is free software; you can redistribute it and/or modify
%  it under the terms of the GNU General Public License as published by
%  the Free Software Foundation; either version 2, or (at your option)
%  any later version.
%
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License for more details.
%
%  You should have received a copy of the GNU General Public License
%  along with this program; see the file COPYING.  If not, write to
%  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
%  Boston, MA 02110-1301, USA.

\begin{code}
{-# OPTIONS -fglasgow-exts #-}
module Darcs.Repository ( Repository, ($-), maybeIdentifyRepository,
                          identifyRepositoryFor,
                          withRepoLock, withRepository, withRepositoryDirectory, withGutsOf,
                    findRepository, amInRepository, amNotInRepository,
                    slurp_pending,
                    slurp_recorded, slurp_recorded_and_unrecorded,
                    withRecorded,
                    get_unrecorded, get_unrecorded_unsorted, get_unrecorded_no_look_for_adds,
                    read_repo, sync_repo,
                    unsafe_fast_read_repo,
                    prefsUrl,
                    read_pending, add_to_pending,
                    tentativelyAddPatch, tentativelyRemovePatches, tentativelyAddToPending,
                    tentativelyReplacePatches,
                    tentativelyMergePatches, considerMergeToWorking,
                    revertRepositoryChanges, finalizeRepositoryChanges,
                    copyRepository, copy_oldrepo_patches,
                    unrevertUrl,
                    applyToWorking, patchSetToPatches,
                    createPristineDirectoryTree, createPartialsPristineDirectoryTree,
                    optimizeInventory, checkPristineAgainstCwd,
                    getMarkedupFile,
                    PatchSet, PatchInfoAnd
                  ) where

import Darcs.Repository.Internal
    (Repository(..), RepoType(..), ($-), pristineFromWorking,
     maybeIdentifyRepository, identifyRepositoryFor,
     findRepository, amInRepository, amNotInRepository,
     slurp_pending,
     slurp_recorded, slurp_recorded_and_unrecorded,
     withRecorded,
     get_unrecorded, get_unrecorded_unsorted, get_unrecorded_no_look_for_adds,
     read_repo, sync_repo,
     unsafe_fast_read_repo,
     prefsUrl, checkPristineAgainstCwd,
     read_pending, add_to_pending,
     withRepoLock, withRepository, withRepositoryDirectory, withGutsOf,
     tentativelyAddPatch, tentativelyRemovePatches, tentativelyAddToPending,
     tentativelyReplacePatches,
     tentativelyMergePatches, considerMergeToWorking,
     revertRepositoryChanges, finalizeRepositoryChanges,
     unrevertUrl,
     applyToWorking, patchSetToPatches,
     createPristineDirectoryTree, createPartialsPristineDirectoryTree,
     optimizeInventory,
     getMarkedupFile,
    )
import Darcs.Repository.Prefs ( unionCaches )
import Darcs.Patch.Set ( PatchSet )

import Control.Monad ( unless, when )

import qualified Darcs.Repository.DarcsRepo as DarcsRepo
import qualified Darcs.Repository.HashedRepo as HashedRepo

import Darcs.Hopefully ( PatchInfoAnd, info )
import Darcs.Repository.Checkpoint ( identify_checkpoint, write_checkpoint_patch, get_checkpoint )
import Darcs.Repository.ApplyPatches ( apply_patches_with_feedback, simple_feedback )
import Darcs.Patch ( RepoPatch, Named, Patch, patch2patchinfo, apply )
import Darcs.Patch.Ordered ( RL(..), mapRL, reverseRL, concatRL )
import Darcs.Repository.Format ( RepoProperty ( HashedInventory ),
                                 format_has )
import Darcs.Patch.Depends ( get_patches_beyond_tag )
import Darcs.Utils ( withCurrentDirectory, catchall )
import Darcs.External ( copyFileOrUrl, Cachable(..) )

#include "impossible.h"

import Darcs.Flags ( DarcsFlag( Verbose, Quiet, Partial ) )
import Printer ( text, putDocLn )
\end{code}

\begin{code}
copyRepository :: RepoPatch p => Repository p -> IO ()
copyRepository fromrepository@(Repo _ opts rf _)
    | Partial `elem` opts && not (format_has HashedInventory rf) =
        do isPartial <- copyPartialRepository fromrepository
           unless (isPartial == IsPartial) $ copyFullRepository fromrepository
    | otherwise = copyFullRepository fromrepository

data PorNP = NotPartial | IsPartial
             deriving ( Eq )

data RepoSort = Hashed | Old

copyInventory :: RepoPatch p => Repository p -> IO ()
copyInventory fromrepo@(Repo fromdir opts rf (DarcsRepository _ cremote)) = do
  Repo todir _ rf2 (DarcsRepository _ c) <- identifyRepositoryFor fromrepo "."
  let copyHashedHashed = HashedRepo.copy_repo (c `unionCaches` cremote) opts fromdir todir
      copyAnythingToOld r = withCurrentDirectory todir $ read_repo r >>=
                            DarcsRepo.write_inventory_and_patches opts
      repoSort rfx | format_has HashedInventory rfx = Hashed
                   | otherwise = Old
  case repoSort rf2 of
    Hashed ->
        if format_has HashedInventory rf
        then copyHashedHashed
        else withCurrentDirectory todir $
             do HashedRepo.revert_tentative_changes
                read_repo fromrepo >>= HashedRepo.write_tentative_inventory c
                HashedRepo.finalize_tentative_changes c opts
    Old -> case repoSort rf of
           Hashed -> copyAnythingToOld fromrepo
           _ -> copy_oldrepo_patches opts fromrepo todir
copyInventory (Repo _ _ _ GitRepository) =
    bug "TODO: need to implement copying of git repositories."

copy_oldrepo_patches :: RepoPatch p => [DarcsFlag] -> Repository p -> FilePath -> IO ()
copy_oldrepo_patches opts repository@(Repo dir _ _ _) out = do
  patches <- DarcsRepo.read_repo opts "."
  mpi <- if Partial `elem` opts
         -- FIXME this should get last pinfo *before*
         -- desired tag...
         then identify_checkpoint repository
         else return Nothing
  DarcsRepo.copy_patches opts dir out $ mapRL info $ since_checkpoint mpi $ concatRL (patches :: PatchSet Patch)
      where --since_checkpoint :: Maybe PatchInfo
            --     -> [PatchInfoAnd] -> [PatchInfoAnd]
            since_checkpoint Nothing ps = ps
            since_checkpoint (Just ch) (hp:<:ps)
                | ch == info hp = hp :<: NilRL
                | otherwise = hp :<: since_checkpoint (Just ch) ps
            since_checkpoint _ NilRL = NilRL

copyPartialRepository :: forall p. RepoPatch p => Repository p -> IO PorNP
copyPartialRepository fromrepository@(Repo _ opts _ _) = do
  mch <- get_checkpoint fromrepository
  case mch :: Maybe (Named p) of
    Nothing -> do putStrLn "No checkpoint."
                  return NotPartial
    Just ch ->
      do copyInventory fromrepository
         withRepoLock opts $- \torepository -> do
           write_checkpoint_patch ch
           local_patches <- unsafe_fast_read_repo torepository
           let pi_ch = patch2patchinfo ch
               needed_patches = reverseRL $ concatRL $
                                get_patches_beyond_tag pi_ch local_patches
           apply opts ch `catch`
                             \e -> fail ("Bad checkpoint!\n" ++ show e)
           apply_patches_with_feedback opts feedback putInfo needed_patches
           putVerbose $ text "Writing the pristine"
           pristineFromWorking torepository
           return IsPartial
      where am_verbose = Verbose `elem` opts
            am_informative = not $ Quiet `elem` opts
            putVerbose s = when am_verbose $ putDocLn s
            putInfo s = when am_informative $ putDocLn s
            feedback = simple_feedback opts "Applying"

copyFullRepository :: RepoPatch p => Repository p -> IO ()
copyFullRepository fromrepository@(Repo fromdir opts rffrom _) = do
  copyInventory fromrepository
  withRepoLock opts $- \torepository@(Repo _ _ rfto _) ->
      if format_has HashedInventory rffrom && format_has HashedInventory rfto
      then do putVerbose $ text "Writing working directory contents..."
              createPristineDirectoryTree torepository "."
              putVerbose $ text "Copying prefs"
              copyFileOrUrl opts (fromdir++"/_darcs/prefs/prefs") "_darcs/prefs/prefs" (MaxAge 600)
                                `catchall` return ()
      else do local_patches <- unsafe_fast_read_repo torepository
              apply_patches_with_feedback opts feedback putInfo $
                                          reverseRL $ concatRL $ local_patches
              putVerbose $ text "Writing the pristine"
              pristineFromWorking torepository
      where putVerbose = when (Verbose `elem` opts) . putDocLn
            putInfo = when (not $ Quiet `elem` opts) . putDocLn
            feedback = simple_feedback opts "Applying"
\end{code}
