%  Copyright (C) 2002-2003 David Roundy
%
%  This program is free software; you can redistribute it and/or modify
%  it under the terms of the GNU General Public License as published by
%  the Free Software Foundation; either version 2, or (at your option)
%  any later version.
%
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License for more details.
%
%  You should have received a copy of the GNU General Public License
%  along with this program; see the file COPYING.  If not, write to
%  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
%  Boston, MA 02110-1301, USA.

\begin{code}
{-# OPTIONS -fffi -fglasgow-exts #-}
#include "gadts.h"
module Darcs.SelectChanges ( with_selected_changes',
                             with_selected_changes_to_files',
                             with_selected_last_changes_to_files',
                             with_selected_changes_reversed',
                             with_selected_last_changes_reversed',
                       with_selected_changes,
                       with_selected_changes_to_files,
                       with_selected_last_changes_to_files,
                       with_selected_changes_reversed,
                       with_selected_last_changes_reversed,
                       view_changes,
                       with_selected_patch_from_repo,
                     ) where
import System.IO
import Data.List ( intersperse )
import Data.Maybe ( isJust, catMaybes )
import Data.Char ( toUpper )
import Control.Monad ( when )
import System ( exitWith, ExitCode(ExitSuccess) )

import Darcs.Hopefully ( PatchInfoAnd, conscientiously )
import Darcs.Repository ( Repository, read_repo, read_pending )
import Darcs.Patch ( RepoPatch, Patchy, Named, Prim, summary, commute,
                     fromPrims, invert, list_touched_files, anonymous )
import qualified Darcs.Patch ( thing, things )
import Darcs.Patch.Ordered ( FL(..), RL(..), (:<)(..), (:>)(..),
                             (+>+), lengthFL, concatRL, mapFL_FL,
                             spanFL, reverseFL, (+<+), mapFL )
import Darcs.Patch.Choices ( PatchChoices, patch_choices, patch_choices_tps,
                             force_first, force_last, make_uncertain, tag,
                      is_patch_first,
                      separate_first_middle_from_last,
                      separate_first_from_middle_last,
                      separate_middle_last_from_first,
                      select_all_middles,
                      separate_last_from_first_middle,
                      force_matching_last,
                      force_matching_first, make_everything_later,
                      TaggedPatch, tp_patch,
                    )
import Darcs.Patch.TouchesFiles ( deselect_not_touching, select_not_touching )
import Darcs.PrintPatch ( printFriendly, printPatch, printPatchPager )
import Darcs.SlurpDirectory ( Slurpy )
import Darcs.Match ( have_nonrange_match, match_a_patch, doesnt_not_match )
import Darcs.Flags ( DarcsFlag( DryRun, All, Summary, DontGrabDeps
                             ) )
import Darcs.Utils ( askUser, promptCharFancy, without_buffering )
import Printer ( text, prefix, ($$), (<>), putDocLn )
#include "impossible.h"
\end{code}

\begin{code}
type WithPatches p a =
        String              -- ^ jobname
     -> [DarcsFlag]         -- ^ opts
     -> Slurpy              -- ^ directory
     -> FL p                -- ^ patches to select among
     -> Maybe Int           -- ^ number of patches (if known)
     -> (FL p :> FL p -> IO a) -- ^ job
     -> IO a                -- ^ result of running job

-- | The only difference with 'WithPatches' is the [FilePath] argument
type WithPatchesToFiles p a =
        String              -- ^ jobname
     -> [DarcsFlag]         -- ^ opts
     -> Slurpy              -- ^ directory
     -> [FilePath]          -- ^ files
     -> FL p                -- ^ patches to select among
     -> Maybe Int           -- ^ number of patches (if known)
     -> (FL p :> FL p -> IO a) -- ^ job
     -> IO a                -- ^ result of running job

with_selected_changes'               :: WithPatches Prim a
with_selected_changes_to_files'      :: WithPatchesToFiles Prim a
with_selected_last_changes_to_files' :: WithPatchesToFiles Prim a
with_selected_changes_reversed'      :: WithPatches Prim a
with_selected_last_changes_reversed' :: WithPatches Prim a
                                     
with_selected_changes'               = wasc  False False False
with_selected_changes_to_files'      = wasc_ False False False
with_selected_last_changes_to_files' = wasc_ True  False False
with_selected_changes_reversed'      = wasc  True  True  False
with_selected_last_changes_reversed' = wasc  False True  False

with_selected_changes               :: RepoPatch p => WithPatches (Named p) a
with_selected_changes_to_files      :: RepoPatch p => WithPatchesToFiles (Named p) a
with_selected_last_changes_to_files :: RepoPatch p => WithPatchesToFiles (Named p) a
with_selected_changes_reversed      :: RepoPatch p => WithPatches (Named p) a
with_selected_last_changes_reversed :: RepoPatch p => WithPatches (Named p) a
                                    
with_selected_changes               = wasc'  False False False
with_selected_changes_to_files      = wasc_' False False False
with_selected_last_changes_to_files = wasc_' True  False False
with_selected_changes_reversed      = wasc'  True  True  False
with_selected_last_changes_reversed = wasc'  False True  False

-- | wasc and wasc_ are just shorthand for with_any_selected_changes
wasc  :: Patchy p => Bool -> Bool -> Bool -> WithPatches p a
wasc v l r j o s = wasc_ v l r j o s []
wasc_ :: Patchy p => Bool -> Bool -> Bool -> WithPatchesToFiles p a
wasc_ = with_any_selected_changes

wasc'  :: RepoPatch p => Bool -> Bool -> Bool -> WithPatches (Named p) a
wasc' v l r j o s = wasc_' v l r j o s []
wasc_' :: RepoPatch p => Bool -> Bool -> Bool -> WithPatchesToFiles (Named p) a
wasc_' = with_any_selected_changes'

with_any_selected_changes :: Patchy p => Bool -> Bool -> Bool -> WithPatchesToFiles p a
with_any_selected_changes isl isr vo jn opts s fs =
    with_any_selected_changes_ isl isr vo (patches_to_consider isl isr fs) jn opts s fs

with_any_selected_changes' :: RepoPatch p => Bool -> Bool -> Bool -> WithPatchesToFiles (Named p) a
with_any_selected_changes' isl isr vo jn opts s fs =
    with_any_selected_changes_ isl isr vo (patches_to_consider' isl isr fs opts) jn opts s fs

view_changes :: RepoPatch p => [DarcsFlag] -> Slurpy -> [FilePath] -> FL (Named p) -> Maybe Int -> IO ()
view_changes opts s fp ps m_ps_len =
    with_any_selected_changes' False False True
                               "view changes" opts s fp ps m_ps_len (\_ -> return ())
\end{code}

\begin{code}
data KeyPress a = KeyPress { kp     :: Char
                           , kpHelp :: String }

helpFor :: String -> [[KeyPress a]] -> String
helpFor jobname options =
  unlines $ [ "How to use "++jobname++":" ]
            ++ (concat $ intersperse [""] $ map (map help) options)
            ++ [ ""
               , "?: show this help"
               , ""
               , "<Space>: accept the current default (which is capitalized)"
               ]
  where help i = kp i:(": "++kpHelp i)

keysFor :: [[KeyPress a]] -> [Char]
keysFor = concatMap (map kp)
\end{code}

\begin{code}
with_selected_patch_from_repo :: RepoPatch p => String -> Repository p -> [DarcsFlag] -> Bool
                              -> (((Named p),[(Named p)]) -> IO ()) -> IO ()
with_selected_patch_from_repo jn repository opts ignore_pending job = do
    p_s <- read_repo repository
    pend <- if ignore_pending
            then return NilFL
            else read_pending repository
    sp <- without_buffering $ wspfr jn (doesnt_not_match opts)
                              (concatRL p_s) [anonymous $ fromPrims pend]
    case sp of
        Just (selected, s_and_pend) ->
          case (last s_and_pend, init s_and_pend) of
          (pend',skipped) ->
            case commute (pend' :> selected) of
            Just (selected' :> _) -> job (selected', skipped)
            Nothing -> impossible
        Nothing -> do putStrLn $ "Cancelling "++jn++" since no patch was selected."
                      exitWith $ ExitSuccess

wspfr :: RepoPatch p => String -> (PatchInfoAnd p -> Bool)
      -> RL (PatchInfoAnd p) -> [(Named p)]
      -> IO (Maybe ((Named p), [(Named p)]))
wspfr _ _ NilRL _ = return Nothing
wspfr jn matches (pihp:<:pps) skipped
    | not $ matches pihp = wspfr jn matches pps (p:skipped)
    | otherwise =
    case commute_by (skipped :< p) of
    Nothing -> do putStr "\nSkipping depended-upon patch:"
                  printFriendly [] p
                  wspfr jn matches pps (p:skipped)
    Just (p' :< skipped') -> do
      printFriendly [] p
      let repeat_this  = wspfr jn matches (pihp:<:pps) skipped
          options = [[ KeyPress 'y' (jn++" this patch")
                     , KeyPress 'n' ("don't "++jn++" it")
                     , KeyPress 'v' "view this patch in full"
                     , KeyPress 'p' "view this patch in full with pager"
                     , KeyPress 'x' "view a summary of this patch"
                     , KeyPress 'q' ("cancel "++jn)
                    ]]
      let prompt  = "Shall I "++jn++" this patch?"
      yorn <- promptCharFancy prompt (keysFor options) (Just 'n') "?h"
      case yorn of
        'y' -> return $ Just (p', skipped')
        'n' -> wspfr jn matches pps (p:skipped)
        'v' -> printPatch p >> repeat_this
        'p' -> printPatchPager p >> repeat_this
        'x' -> do putDocLn $ prefix "    " $ summary p
                  repeat_this
        'q' -> do putStrLn $ jn_cap++" cancelled."
                  exitWith $ ExitSuccess
        _   -> do putStrLn $ helpFor jn options
                  repeat_this
  where p = conscientiously (\e -> text ""
                             $$ text ("Can't " ++ jn ++ " patch")
                             $$ e
                             $$ text "since I can't read it."
                             $$ text ""
                             $$ text "The most likely reason is that you are using a 'partial'"
                             <> text "repository that doesn't contain this patch.") pihp
        jn_cap = (toUpper $ head jn) : tail jn

commute_by :: Patchy p => [p] :< p -> Maybe (p :< [p])
commute_by ([] :< a) = Just (a :< [])
commute_by (p:ps :< a) =
    case commute (a :> p) of
    Nothing -> Nothing
    Just (p':>a') -> case commute_by (ps :< a') of
                     Nothing -> Nothing
                     Just (a'' :< ps') -> Just (a'' :< p':ps')
\end{code}

\begin{code}
with_any_selected_changes_ :: Patchy p => Bool -- ^ is last
                           -> Bool -- ^ is reversed
                           -> Bool -- ^ view only
                           -> (FL p -> FL p :> FL p)
                           -> WithPatchesToFiles p a
with_any_selected_changes_ islast isreversed viewonly p2c
                          jobname opts _ fs ps m_ps_len job =
 if All `elem` opts || DryRun `elem` opts
 then job $ if isreversed then invert other_ps :> invert ps_to_consider
                          else ps_to_consider :> other_ps
 else
    do pc <- without_buffering $
             if not viewonly
             then tentatively_text_select jobname islast isreversed
                                     opts ps_len ps_len_calced 0
                                     NilRL init_tps init_pc
             else text_view jobname islast isreversed opts ps_len
                            ps_len_calced 0 NilRL init_tps init_pc
       job $ selected_patches islast isreversed other_ps pc
   where ps_to_consider :> other_ps = p2c ps
         (init_pc, init_tps) = patch_choices_tps $ if islast then other_ps
                                                             else ps_to_consider
         (ps_len, ps_len_calced)
           = if null fs && not (have_nonrange_match opts)
             then case m_ps_len of
                  Just l  -> (l         , True)
                  Nothing -> (lengthFL ps , False)
             else (lengthFL ps_to_consider, False)

patches_to_consider :: Patchy p => Bool        -- ^ is last
                    -> Bool        -- ^ is reversed
                    -> [FilePath]  -- ^ files
                    -> FL p     -- ^ patches
                    -> (FL p :> FL p)
patches_to_consider islast isreversed fs ps =
  let ps' = if isreversed then invert ps else ps
      f = if islast && not isreversed
          then separate_middle_last_from_first
          else separate_first_middle_from_last
      deal_with_fs = if islast
                     then select_not_touching   fs
                     else deselect_not_touching fs
  in if null fs then if islast then NilFL :> ps'  else ps' :> NilFL
                else tp_patches $ f $ deal_with_fs $ patch_choices ps'

patches_to_consider' :: RepoPatch p => Bool        -- ^ is last
                     -> Bool        -- ^ is reversed
                     -> [FilePath]  -- ^ files
                     -> [DarcsFlag] -- ^ opts
                     -> FL (Named p) -- ^ patches
                     -> FL (Named p) :> FL (Named p)
patches_to_consider' islast isreversed fs opts ps =
  let ps' = if isreversed then invert ps else ps
      f = if islast && not isreversed
          then separate_middle_last_from_first
          else separate_first_middle_from_last
      deal_with_fs = if islast
                     then select_not_touching   fs
                     else deselect_not_touching fs
      deselect_unwanted pc =
        if have_nonrange_match opts
        then if islast
             then bug "don't support patch matching with islast in wasp"
             else if DontGrabDeps `elem` opts
                  then force_matching_last (not.iswanted) pc
                  else make_everything_later $ force_matching_first iswanted pc
        else pc
      iswanted = let maybe_invert = if isreversed then invert else id
                 in (match_a_patch opts . maybe_invert . tp_patch)
  in if null fs && not (have_nonrange_match opts)
     then if islast then NilFL :> ps'  else ps' :> NilFL
     else tp_patches $ f $ deal_with_fs $ deselect_unwanted $ patch_choices ps'

-- | Returns the results of a patch selection user interaction
selected_patches :: Patchy p => Bool -> Bool
                 -> FL p -- ^ patches set aside
                 -> PatchChoices p
                 -> (FL p :> FL p)
selected_patches islast isreversed other_ps pc =
  case (islast, isreversed) of
  (True, False) ->
      case separate_last_from_first_middle pc of
      xs :> ys -> other_ps +>+ mapFL_FL tp_patch xs :> mapFL_FL tp_patch ys
  (False, False) ->
      case separate_first_from_middle_last pc of
      xs :> ys -> mapFL_FL tp_patch xs :> mapFL_FL tp_patch ys +>+ other_ps
  (True, True) ->
      case separate_first_middle_from_last pc of
      xs :> ys -> lengthFL xs `seq` -- Aaack, ugly hack here...
                  invert (mapFL_FL tp_patch ys) :> invert (other_ps +>+ mapFL_FL tp_patch xs)
  (False, True) ->
      case separate_first_from_middle_last pc of
      xs :> ys -> invert (mapFL_FL tp_patch ys +>+ other_ps) :> invert (mapFL_FL tp_patch xs)

text_select :: Patchy p => String -> Bool -> Bool -> [DarcsFlag] -> Int -> Bool -> Int
            -> RL (TaggedPatch p) -> FL (TaggedPatch p) -> PatchChoices p
            -> IO (PatchChoices p)

text_select _ _ _ _ _ _ _ _ NilFL pc = return pc
text_select jn islast isinverted opts n_max n_max_calced n
            tps_done tps_todo@(tp:>:tps_todo') pc = do
      printFriendly opts viewp
      repeat_this -- prompt the user
    where
        do_next = tentatively_text_select jn islast isinverted opts
                                          n_max n_max_calced
                                          (n+1) (tp:<:tps_done) tps_todo'
        helper :: PatchChoices p -> p
        helper = undefined
        thing  = Darcs.Patch.thing (helper pc)
        things = Darcs.Patch.things (helper pc)
        options_basic =
           [ KeyPress 'y' (jn++" this "++thing)
           , KeyPress 'n' ("don't "++jn++" it")
           , KeyPress 'w' ("wait and decide later, defaulting to no") ]
        options_file =
           [ KeyPress 's' ("don't "++jn++" the rest of the changes to this file")
           , KeyPress 'f' (jn++" the rest of the changes to this file") ]
        options_view =
           [ KeyPress 'v' ("view this "++thing++" in full")
           , KeyPress 'p' ("view this "++thing++" in full with pager") ]
        options_summary =
           [ KeyPress 'x' ("view a summary of this "++thing) ]
        options_quit =
           [ KeyPress 'd' (jn++" selected "++things++", skipping all the remaining "++things)
           , KeyPress 'a' (jn++" all the remaining "++things)
           , KeyPress 'q' ("cancel "++jn) ]
        options_nav =
           [ KeyPress 'j' ("skip to next "++thing)
           , KeyPress 'k' ("back up to previous "++thing) ]
        options_count =
           [ KeyPress 'c' ("calculate number of "++things) ]
        options = [options_basic]
                  ++ (if is_single_file_patch then [options_file] else [])
                  ++ [options_view ++
                      if Summary `elem` opts then [] else options_summary]
                  ++ [options_quit]
                  ++ [options_nav ++
                      if n_max_calced then [] else options_count]
        prompt = "Shall I "++jn++" this "++thing++"? "
               ++ "(" ++ show (n+1) ++ "/"
               ++ (if n_max_calced then show n_max else "?") ++ ") "
        repeat_this = do
          yorn <- promptCharFancy prompt (keysFor options) (Just the_default) "?h"
          case yorn of
            'y' -> do_next $ force_yes (tag tp) pc
            'n' -> do_next $ force_no (tag tp) pc
            'w' -> do_next $ make_uncertain (tag tp) pc
            's' -> do_next $ skip_file
            'f' -> do_next $ do_file
            'v' -> printPatch viewp >> repeat_this
            'p' -> printPatchPager viewp >> repeat_this
            'x' -> do putDocLn $ prefix "    " $ summary viewp
                      repeat_this
            'd' -> return pc
            'a' -> do ask_confirmation
                      return $ select_all_middles islast pc
            'q' -> do putStrLn $ jn_cap++" cancelled."
                      exitWith $ ExitSuccess
            'j' -> case tps_todo' of
                       NilFL -> -- May as well work out the length now we have all
                                -- the patches in memory
                                text_select jn islast isinverted opts
                                    n_max True n tps_done tps_todo pc
                       _ -> text_select jn islast isinverted opts
                                n_max n_max_calced (n+1) (tp:<:tps_done) tps_todo' pc
            'k' -> case tps_done of
                        NilRL -> repeat_this
                        (tp':<:tps_done') ->
                           text_select jn islast isinverted opts
                               n_max n_max_calced (n-1) tps_done' (tp':>:tps_todo) pc
            'c' -> text_select jn islast isinverted opts
                                        n_max True n tps_done tps_todo pc
            _   -> do putStrLn $ helpFor jn options
                      repeat_this
        force_yes = if islast then force_last else force_first
        force_no  = if islast then force_first else force_last
        patches_to_skip = (tag tp:) $ catMaybes
                        $ mapFL (\tp' -> if list_touched_files tp' == touched_files
                                         then Just (tag tp')
                                         else Nothing) tps_todo'
        skip_file = foldr force_no pc patches_to_skip
        do_file = foldr force_yes pc patches_to_skip
        the_default = get_default islast $ is_patch_first tp pc
        jn_cap = (toUpper $ head jn) : tail jn
        touched_files = list_touched_files $ tp_patch tp
        is_single_file_patch = length touched_files == 1
        viewp = if isinverted then invert (tp_patch tp) else tp_patch tp
        ask_confirmation =
            if jn `elem` ["unpull", "unrecord", "obliterate"]
            then do yorn <- askUser $ "Really " ++ jn ++ " all undecided patches? "
                    case yorn of
                     ('y':_) -> return ()
                     _ -> exitWith $ ExitSuccess
            else return ()

text_view :: Patchy p => String -> Bool -> Bool -> [DarcsFlag] -> Int -> Bool -> Int
            -> RL (TaggedPatch p) -> FL (TaggedPatch p) -> PatchChoices p
            -> IO (PatchChoices p)

text_view _ _ _ _ _ _ _ _ NilFL _ = return $ patch_choices NilFL --return pc
text_view jn islast isinverted opts n_max n_max_calced n
            tps_done tps_todo@(tp:>:tps_todo') pc = do
      printFriendly opts viewp
      putStr "\n"
      repeat_this -- prompt the user
    where
        prev_patch = case tps_done of
                       NilRL -> repeat_this
                       (tp':<:tps_done') ->
                         text_view jn islast isinverted opts
                            n_max n_max_calced (n-1) tps_done' (tp':>:tps_todo) pc
        next_patch = case tps_todo' of
                         NilFL -> -- May as well work out the length now we have all
                                  -- the patches in memory
                               text_view jn islast isinverted opts n_max
                                   True n tps_done tps_todo' pc
                         _ -> text_view jn islast isinverted opts n_max
                                  n_max_calced (n+1) (tp:<:tps_done) tps_todo' pc
        options_yn =
          [ KeyPress 'y' "skip to the next patch"
          , KeyPress 'n' ("quit "++jn) ]
        options_view =
          [ KeyPress 'v' "view this patch in full"
          , KeyPress 'p' "view this patch in fill with pager" ]
        options_summary =
          [ KeyPress 'x' "view a summary of this patch" ]
        options_nav =
          [ KeyPress 'q' ("quit "++jn)
          , KeyPress 'k' "back up to previous patch"
          , KeyPress 'j' "skip to next patch" ]
        options_count =
          [ KeyPress 'c' "calculate number of patches" ]
        options = [ options_yn ]
                  ++ [ options_view ++
                       if Summary `elem` opts then [] else options_summary ]
                  ++ [ options_nav ++
                       if n_max_calced then [] else options_count ]
        prompt = "Shall I continue to "++jn++"? "
               ++ "(" ++ show (n+1) ++ "/"
               ++ (if n_max_calced then show n_max else "?") ++ ")"
        repeat_this = do
          yorn <- promptCharFancy prompt (keysFor options) (Just 'y') "?h"
          case yorn of
            'y' -> next_patch
            'n' -> exitWith ExitSuccess
            'v' -> printPatch viewp >> repeat_this
            'p' -> printPatchPager viewp >> repeat_this
            'x' -> do putDocLn $ prefix "    " $ summary viewp
                      repeat_this
            'q' -> exitWith ExitSuccess
            'k' -> prev_patch
            'j' -> next_patch
            'c' -> text_view jn islast isinverted opts
                       n_max True n tps_done tps_todo pc
            _   -> do putStrLn $ helpFor jn options
                      repeat_this
        viewp = if isinverted then invert (tp_patch tp) else tp_patch tp

tentatively_text_select :: Patchy p => String -> Bool -> Bool -> [DarcsFlag]
                        -> Int -> Bool -> Int -> RL (TaggedPatch p) -> FL (TaggedPatch p)
                        -> PatchChoices p
                        -> IO (PatchChoices p)
tentatively_text_select _ _ _ _ _ _ _ _ NilFL pc = return pc
tentatively_text_select jn islast isinverted opts n_max n_max_calced n ps_done
                        ps_todo pc = do
  when (numSkipped > 0) $
    putStrLn ("Skipped " ++ show numSkipped ++ _patch_ ++ " due to dependencies.")
  text_select jn islast isinverted opts n_max n_max_calced (n + numSkipped)
    (reverseFL skipped +<+ ps_done) unskipped pc
 where
  skipped :> unskipped = spanFL (\p -> isJust $ is_patch_first p pc) ps_todo
  numSkipped  = lengthFL skipped
  _patch_ = if numSkipped > 1 then " patches" else " patch"

get_default :: Bool -> Maybe Bool -> Char
get_default _ Nothing = 'w'
get_default True (Just True) = 'n'
get_default True (Just False) = 'y'
get_default False (Just True) = 'y'
get_default False (Just False) = 'n'
\end{code}

\begin{code}
tp_patches :: (FL (TaggedPatch p) :> FL (TaggedPatch p)) C(x y)
           -> (FL p :> FL p) C(x y)
tp_patches (x:>y) = mapFL_FL tp_patch x :> mapFL_FL tp_patch y
\end{code}
